/*-
 * Copyright (c) 2004
 *     Michael B. Durian
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 4. The name of the Author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#ifdef linux
#include <sys/ioctl.h>
#endif

#include "intrepidcs.h"

#define MAX_PACKET_SIZE	(ICS_MAX_DATA_SIZE + 3)

struct ics_interface_type {
	speed_t	baud;
	const	char *name;
	double	timestamp_tick;
	double	timestamp_overflow;
};

#ifdef linux
static const struct ics_interface_type Interfaces[] = {
	{B2000000, "ValueCAN", 0.000001, 0.065536}, //Linux needs these constants for cfsetispeed(3) and cfsetospeed(3)
	{B3000000, "neoVI", 0.0000016, 0.1048576}
};
#else
static const struct ics_interface_type Interfaces[] = {
	{2000000, "ValueCAN", 0.000001, 0.065536},
	{3000000, "neoVI", 0.0000016, 0.1048576}
};
#endif
static const int NumInterfaces = sizeof(Interfaces) / sizeof(Interfaces[0]);

static ssize_t interface_write(struct ics_interface *iface, uint8_t *ptr,
    size_t nbytes);
static ssize_t interface_read(struct ics_interface *iface, uint8_t *ptr,
    size_t nbytes);
static size_t interface_read_from_buffer(struct ics_interface *iface,
    uint8_t *buf, size_t nbytes);
static const char *get_net_id_str(uint8_t id);
static int stream_allow_can_msg(struct ics_stream_state *ss,
    struct ics_can_msg *msg);

#define MAX_ERR_LEN	128
#define BUFFER_SIZE	1024
struct ics_interface {
	int	index;
	int	fd;
	uint8_t	buf[BUFFER_SIZE];
	uint8_t	*drain;
	uint8_t	*fill;
	int	count;
	char	err[MAX_ERR_LEN];
};

struct can_filter {
	uint32_t	mask;
	uint32_t	id;
	uint32_t	masked_id;
	struct	can_filter *next;
	struct	can_filter *prev;
};

struct ics_stream_state {
	double	last_timestamp;
	double	overflow;

	double	timestamp_tick;
	double	timestamp_overflow;

	struct	can_filter *filters;

	/* callback defintions */
	ics_device_status_func	*dev_stat_cb;
	void	*dev_stat_arg;

	ics_rx_overflow_func	*rx_overflow_cb;
	void 	*rx_overflow_arg;

	ics_tx_complete_func	*tx_complete_cb;
	void	*tx_complete_arg;

	ics_can_error_func	*can_error_cb;
	void 	*can_error_arg;

	ics_pic_overflow_func	*pic_overflow_cb;
	void	*pic_overflow_arg;

	ics_main51_overflow_func	*main51_overflow_cb;
	void	*main51_overflow_arg;

	ics_hscan_func	*hscan_cb;
	void	*hscan_arg;

	ics_swcan_func  *swcan_cb;
	void    *swcan_arg;

	ics_mscan_func  *mscan_cb;
	void    *mscan_arg;

	ics_lsftcan_func  *lsftcan_cb;
	void    *lsftcan_arg;
};

int
ics_interface_write_packet(struct ics_interface *iface, struct ics_packet *p)
{
	int i, num_written, num_to_write, total_written;
	uint8_t cksum, packet[MAX_PACKET_SIZE], *pptr;

	if (p->target >= 16) {
		snprintf(iface->err, MAX_ERR_LEN, "invalid target");
		return (0);
	}
	if (p->len == 0) {
		snprintf(iface->err, MAX_ERR_LEN, "packet has no data");
		return (0);
	}
	if (p->len > ICS_MAX_DATA_SIZE) {
		snprintf(iface->err, MAX_ERR_LEN, "packet has too much data "
		    "(%d >= %d)", p->len, ICS_MAX_DATA_SIZE);
		return (0);
	}

#ifdef PRINT_SENT_PACKETS
	/**
	 * XXX: debugging purposes
	 */
	char output[100];
	ics_packet_format(output, 100, p);
	printf("Sending packet: %s\n", output);
#endif


	pptr = packet;
	*pptr++ = 0xaa;
	cksum = (p->len << 4) | p->target;
	*pptr++ = cksum;
	for (i = 0; i < p->len; i++) {
		*pptr++ = p->data[i];
		cksum += p->data[i];
	}
	cksum = ~cksum;
	cksum++;
	*pptr = cksum;

	total_written = 0;
	while (total_written < p->len + 3) {
		num_to_write = (p->len + 3 - total_written);
		num_written = interface_write(iface, &packet[total_written],
		    num_to_write);
		if (num_written == -1) {
			snprintf(iface->err, MAX_ERR_LEN, "write failed: %s",
			    strerror(errno));
			return (0);
		}
		total_written += num_written;
	}
	return (1);
}

int
ics_interface_read_packet(struct ics_interface *iface, struct ics_packet *p,
    int *is_valid)
{
	int i, num_read, total_read;
	uint8_t calcsum, cksum, start, lentgt;

	/* assume a bad message */
	*is_valid = 0;

	if (interface_read(iface, &start, 1) == -1) {
		snprintf(iface->err, MAX_ERR_LEN, "read start failed: %s",
		    strerror(errno));
		return (0);
	}
	if (start == 0xaa) {
		if (interface_read(iface, &lentgt, 1) == -1) {
			snprintf(iface->err, MAX_ERR_LEN, "read len/tgt "
			    "failed: %s", strerror(errno));
			return (0);
		}
		p->len = (lentgt >> 4) & 0x0f;
		p->target = lentgt & 0x0f;

		total_read = 0;
		while (total_read < p->len) {
			num_read = interface_read(iface, &p->data[total_read],
			    p->len);
			if (num_read == -1) {
				snprintf(iface->err, MAX_ERR_LEN, "read data "
				    "failed: %s", strerror(errno));
				return (0);
			}
			total_read += num_read;
		}

		if (interface_read(iface, &cksum, 1) == -1) {
			snprintf(iface->err, MAX_ERR_LEN, "read cksum "
			    "failed: %s", strerror(errno));
			return (0);
		}

		calcsum = lentgt;
		for (i = 0; i < p->len; i++)
			calcsum += p->data[i];
		calcsum = ~calcsum;
		calcsum++;

		/* if the checksum matches, we have a good message */
		if (calcsum == cksum)
			*is_valid = 1;
	}
#ifdef PRINT_RECEIVED_PACKETS
	/**
	 * XXX: debugging purposes
	 */
	char output[100];
	ics_packet_format(output, 100, p);
	printf("Received packet: %s\n", output);
#endif
	return (1);
}

void
ics_packet_format(char *str, int str_len, struct ics_packet *p)
{
	int i, slen;
	uint8_t cksum;

	slen = 0;
	slen += snprintf(&str[slen], str_len, "%s:",
	    get_net_id_str(p->target));
	for (i = 0; i < p->len; i++)
		slen += snprintf(&str[slen], str_len, " %02x", p->data[i]);

	/* recalculate the checksum so we can print it */
	cksum = (p->len << 4) | p->target;
	for (i = 0; i < p->len; i++)
		cksum += p->data[i];
	cksum = ~cksum;
	cksum++;

	slen += snprintf(&str[slen], str_len, " (aa %02x",
	    (p->len << 4) | p->target);
	for (i = 0; i < p->len; i++)
		slen += snprintf(&str[slen], str_len, " %02x", p->data[i]);
	snprintf(&str[slen], str_len, " %02x)", cksum);
}

int
ics_packet_unpack_can_msg(struct ics_packet *p, struct ics_can_msg *msg)
{
	int i, ret;

	if (!(p->len >= 5 && p->len <= 15))
		ret = 0;
	else {
		msg->netid = p->target;

		msg->timestamp = (p->data[0] << 8) | p->data[1];

		/* determine if it is a standard ID or and extended ID */
		if (p->data[3] & 0x08) {
			/* extended frame */
			msg->is_extended_id = 1;

			msg->can_id = ((uint32_t)p->data[2] << 21) |
			    (((uint32_t)p->data[3] & 0xe0) << 13) |
			    (((uint32_t)p->data[3] & 0x03) << 16) |
			    ((uint32_t)p->data[4] << 8) |
			    p->data[5];

			if (p->data[6] & 0x40)
				msg->is_remote_frame = 1;
			else
				msg->is_remote_frame = 0;

			msg->data_len = p->data[6] & 0x0f;

			for (i = 0; i < msg->data_len; i++)
				msg->data[i] = p->data[7 + i];
		} else {
			msg->is_extended_id = 0;

			msg->can_id = ((uint32_t)p->data[2] << 3) |
			    (((uint32_t)p->data[3] & 0xe0) >> 5);

			if (p->data[3] & 0x10)
				msg->is_remote_frame = 1;
			else
				msg->is_remote_frame = 0;

			msg->data_len = p->data[4] & 0x0f;

			for (i = 0; i < msg->data_len; i++)
				msg->data[i] = p->data[5 + i];
		}
		ret = 1;
	}
	return (ret);
}

struct ics_stream_state *
ics_stream_init(struct ics_interface *iface, char *err, int err_len)
{
	struct ics_stream_state *ss;

	ss = malloc(sizeof(struct ics_stream_state));
	if (ss == NULL) {
		snprintf(err, err_len, "Out of memory");
		return (NULL);
	}

	ss->timestamp_tick = Interfaces[iface->index].timestamp_tick;
	ss->timestamp_overflow = Interfaces[iface->index].timestamp_overflow;

	/* Sorry, but we can't really count on memset(3) being available since
	 * it's not ANSI standard */
	ss->filters = NULL;

	ss->dev_stat_cb = NULL;
	ss->dev_stat_arg = NULL;
	ss->rx_overflow_cb = NULL;
	ss->rx_overflow_arg = NULL;
	ss->tx_complete_cb = NULL;
	ss->tx_complete_arg = NULL;
	ss->can_error_cb = NULL;
	ss->can_error_arg = NULL;
	ss->pic_overflow_cb = NULL;
	ss->pic_overflow_arg = NULL;
	ss->main51_overflow_cb = NULL;
	ss->main51_overflow_arg = NULL;
	ss->hscan_cb = NULL;
	ss->hscan_arg = NULL;
	ss->mscan_cb = NULL;
	ss->mscan_arg = NULL;
	ss->swcan_cb = NULL;
	ss->swcan_arg = NULL;
	ss->lsftcan_cb = NULL;
	ss->lsftcan_arg = NULL;

	/* start timing from 0 */
	ss->last_timestamp = 0;
	ss->overflow = 0;

	return (ss);
}

void
ics_stream_release(struct ics_stream_state *ss)
{

	free(ss);
}

void *
ics_stream_add_filter(struct ics_stream_state *ss, uint32_t mask, uint32_t id)
{
	struct can_filter *f, *fptr;

	f = malloc(sizeof(struct can_filter));
	if (f == NULL)
		return (NULL);
	f->mask = mask;
	f->id = id;
	/* compute this now to save time later */
	f->masked_id = f->mask & f->id;
	f->next = NULL;
	f->prev = NULL;

	if (ss->filters == NULL) {
		ss->filters = f;
	} else {
		fptr = ss->filters;
		while (fptr->next != NULL)
			fptr = fptr->next;
		fptr->next = f;
		f->prev = fptr;
	}
	return (f);
}

void
ics_stream_remove_filter(struct ics_stream_state *ss, void *filter_id)
{
	struct can_filter *f;

	f = (struct can_filter *)filter_id;

	if (f->prev != NULL)
		f->prev->next = f->next;
	if (f->next != NULL)
		f->next->prev = f->prev;
	if (ss->filters == f)
		ss->filters = f->next;
	free(f);
}

int
ics_stream_process_packet(struct ics_stream_state *ss, struct ics_packet *p)
{
	struct ics_dev_status dev_stat;
	struct ics_rx_overflow rx_overflow;
	struct ics_tx_complete tx_complete;
	struct ics_can_error can_error;
	struct ics_main51_overflow main51_overflow;
	struct ics_can_msg can_msg;
	int packet_processed;

	packet_processed = 0;

	/* determine the message type */
	if (p->target == ICS_NETID_DEVICE) {
		if (p->len >= 1) {
			switch (p->data[0]) {
			case ICS_MSG_TYPE_DEV_STATUS:
				/* timer overflow */
				if (ss->dev_stat_cb != NULL) {
					dev_stat.sync_count = p->data[1];
					dev_stat.reset_status = p->data[2];
					ss->dev_stat_cb(ss, &dev_stat,
					    ss->dev_stat_arg);
				}
				packet_processed = 1;
				break;
			case ICS_MSG_TYPE_DEV_RX_OVFL:
				if (ss->rx_overflow_cb != NULL) {
					rx_overflow.netid = p->data[1];
					rx_overflow.timestamp =
					    (p->data[2] << 8) | p->data[3];
					ss->rx_overflow_cb(ss, &rx_overflow,
					    ss->rx_overflow_arg);
				}
				packet_processed = 1;
				break;
			case ICS_MSG_TYPE_DEV_TX_COMPL:
				if (ss->tx_complete_cb != NULL) {
					tx_complete.netid = p->data[1];
					tx_complete.timestamp =
					    (p->data[2] << 8) | p->data[3];
					tx_complete.desc = (p->data[4] << 8) |
					    p->data[5];
					ss->tx_complete_cb(ss, &tx_complete,
					    ss->tx_complete_arg);
				}
				packet_processed = 1;
				break;
			case ICS_MSG_TYPE_DEV_CAN_ERROR:
				if (ss->can_error_cb != NULL) {
					can_error.netid = p->data[1];
					can_error.timestamp =
					    (p->data[2] << 8) | p->data[3];
					if (p->data[4] & (1 << 7))
						can_error.rx1ovr = 1;
					else
						can_error.rx1ovr = 0;
					if (p->data[4] & (1 << 6))
						can_error.rx0ovr = 1;
					else
						can_error.rx0ovr = 0;
					if (p->data[4] & (1 << 5))
						can_error.txbo = 1;
					else
						can_error.txbo = 0;
					if (p->data[4] & (1 << 4))
						can_error.txep = 1;
					else
						can_error.txep = 0;
					if (p->data[4] & (1 << 3))
						can_error.rxep = 1;
					else
						can_error.rxep = 0;
					if (p->data[4] & (1 << 2))
						can_error.txwar = 1;
					else
						can_error.txwar = 0;
					if (p->data[4] & (1 << 1))
						can_error.rxwar = 1;
					else
						can_error.rxwar = 0;
					if (p->data[4] & (1 << 0))
						can_error.ewarn = 1;
					else
						can_error.ewarn = 0;
					ss->can_error_cb(ss, &can_error,
					    ss->can_error_arg);
				}
				packet_processed = 1;
				break;
			case ICS_MSG_TYPE_DEV_PIC_OVFL:
				if (ss->pic_overflow_cb != NULL)
					ss->pic_overflow_cb(ss,
					    ss->pic_overflow_arg);
				packet_processed = 1;
				break;
			}
		}

		/* Whoops! this will pick up TX complete packets as well if
		 * packet_processed is not checked */
		if (p->len >= 1 && !packet_processed) {
			switch (p->data[0]) {
			case ICS_MSG_TYPE_MAIN51_TX_OVFL:
				if (ss->main51_overflow_cb != NULL) {
					main51_overflow.netid = p->data[1];
					ss->main51_overflow_cb(ss,
					    &main51_overflow,
					    ss->main51_overflow_arg);
				}
				packet_processed = 1;
			}
		}
	} else if (p->target == ICS_NETID_HSCAN) {
		if (ss->hscan_cb != NULL) {
			if (!ics_packet_unpack_can_msg(p, &can_msg))
				ss->hscan_cb(ss, NULL, ss->hscan_arg);
			else {
				if (stream_allow_can_msg(ss, &can_msg))
					ss->hscan_cb(ss, &can_msg,
					    ss->hscan_arg);
			}
		}
		packet_processed = 1;
	} else if (p->target == ICS_NETID_SWCAN) {
		if (ss->swcan_cb != NULL) {
			if (!ics_packet_unpack_can_msg(p, &can_msg))
				ss->swcan_cb(ss, NULL, ss->hscan_arg);
			else {
				if (stream_allow_can_msg(ss, &can_msg))
					ss->swcan_cb(ss, &can_msg,
					    ss->swcan_arg);
			}
		}
		packet_processed = 1;
	} else if (p->target == ICS_NETID_MSCAN) {
		if (ss->mscan_cb != NULL) {
			if (!ics_packet_unpack_can_msg(p, &can_msg))
				ss->mscan_cb(ss, NULL, ss->mscan_arg);
			else {
				if (stream_allow_can_msg(ss, &can_msg))
					ss->mscan_cb(ss, &can_msg,
					    ss->mscan_arg);
			}
		}
		packet_processed = 1;
	} else if (p->target == ICS_NETID_LSFTCAN) {
		if (ss->lsftcan_cb != NULL) {
			if (!ics_packet_unpack_can_msg(p, &can_msg))
				ss->lsftcan_cb(ss, NULL, ss->lsftcan_arg);
			else {
				if (stream_allow_can_msg(ss, &can_msg))
					ss->lsftcan_cb(ss, &can_msg,
					    ss->lsftcan_arg);
			}
		}
		packet_processed = 1;
	}


	return (packet_processed);
}

void
ics_stream_set_device_status_callback(struct ics_stream_state *ss,
    ics_device_status_func *func, void *client_data)
{

	ss->dev_stat_cb = func;
	ss->dev_stat_arg = client_data;
}

void
ics_stream_set_rx_overflow_callback(struct ics_stream_state *ss,
    ics_rx_overflow_func *func, void *client_data)
{

	ss->rx_overflow_cb = func;
	ss->rx_overflow_arg = client_data;
}

void
ics_stream_set_tx_complete_callback(struct ics_stream_state *ss,
    ics_tx_complete_func *func, void *client_data)
{

	ss->tx_complete_cb = func;
	ss->tx_complete_arg = client_data;
}

void
ics_stream_set_can_error_callback(struct ics_stream_state *ss,
    ics_can_error_func *func, void *client_data)
{

	ss->can_error_cb = func;
	ss->can_error_arg = client_data;
}

void
ics_stream_set_pic_overflow_callback(struct ics_stream_state *ss,
    ics_pic_overflow_func *func, void *client_data)
{

	ss->pic_overflow_cb = func;
	ss->pic_overflow_arg = client_data;
}

void
ics_stream_set_main51_overflow_callback(struct ics_stream_state *ss,
    ics_main51_overflow_func *func, void *client_data)
{

	ss->main51_overflow_cb = func;
	ss->main51_overflow_arg = client_data;
}

void
ics_stream_set_hscan_callback(struct ics_stream_state *ss,
    ics_hscan_func *func, void *client_data)
{

	ss->hscan_cb = func;
	ss->hscan_arg = client_data;
}

void
ics_stream_set_swcan_callback(struct ics_stream_state *ss,
    ics_swcan_func *func, void *client_data)
{

	ss->swcan_cb = func;
	ss->swcan_arg = client_data;
}

void
ics_stream_set_mscan_callback(struct ics_stream_state *ss,
    ics_mscan_func *func, void *client_data)
{

	ss->mscan_cb = func;
	ss->mscan_arg = client_data;
}

void
ics_stream_set_lsftcan_callback(struct ics_stream_state *ss,
    ics_lsftcan_func *func, void *client_data)
{

	ss->lsftcan_cb = func;
	ss->lsftcan_arg = client_data;
}

double
ics_stream_get_timestamp_tick(struct ics_stream_state *ss)
{

	return (ss->timestamp_tick);
}

double
ics_stream_get_timestamp_overflow(struct ics_stream_state *ss)
{

	return (ss->timestamp_overflow);
}

void
ics_stream_update_device_status(struct ics_stream_state *ss,
    struct ics_dev_status *ds)
{

	/*
	 * XXX
	 * Should check sync_count to verify we haven't missed any
	 * status packets and update overflow by more than one
	 * timestamp_overflow if a packet was determined to be dropped.
	 */
	ss->overflow += ss->timestamp_overflow;
}

void
ics_stream_format_can_msg(struct ics_stream_state *ss, char *msg_text,
    int max_text, struct ics_can_msg *msg, int delta)
{
	double new_timestamp;
	int i, slen;

	slen = 0;
	new_timestamp = msg->timestamp * ss->timestamp_tick + ss->overflow;
	if (delta)
		slen += snprintf(&msg_text[slen], max_text, "%0.6f",
		    new_timestamp - ss->last_timestamp);
	else
		slen += snprintf(&msg_text[slen], max_text, "%0.6f",
		    new_timestamp);
	ss->last_timestamp = new_timestamp;
	if (msg->is_extended_id)
		slen += snprintf(&msg_text[slen], max_text, " %08x",
		    msg->can_id);
	else
		slen += snprintf(&msg_text[slen], max_text, " %03x",
		    msg->can_id);
	if (msg->is_remote_frame)
		slen += snprintf(&msg_text[slen], max_text, " remote");
	slen += snprintf(&msg_text[slen], max_text, ":");
	for (i = 0; i < msg->data_len; i++)
		slen += snprintf(&msg_text[slen], max_text, " %02x",
		    msg->data[i]);
}

void
ics_packet_create_can_enable(struct ics_packet *p)
{

	p->target = ICS_NETID_MAIN51;
	p->len = 2;
	p->data[0] = 0x07;
	p->data[1] = 0x01;
}

void
ics_packet_create_can_disable(struct ics_packet *p)
{

	p->target = ICS_NETID_MAIN51;
	p->len = 2;
	p->data[0] = 0x07;
	p->data[1] = 0x00;
}

void
ics_packet_create_can_msg(struct ics_packet *p, struct ics_can_msg *msg)
{
	int i;

	p->target = msg->netid;

	/* XXX description ID? */
	p->data[0] = (msg->timestamp >> 8) & 0xff;
	p->data[1] = msg->timestamp & 0xff;

	if (msg->is_extended_id) {
		p->data[2] = msg->can_id >> 21;
		p->data[3] = (((msg->can_id & 0x001c0000) >> 13) & 0xff) |
		    (((msg->can_id & 0x00030000) >> 16) & 0xff) | 0x08;
		p->data[4] = (msg->can_id >> 8) & 0xff;
		p->data[5] = msg->can_id & 0xff;

		p->data[6] = 0;
		if (msg->is_remote_frame)
			p->data[6] |= 0x40;
		p->data[6] |= msg->data_len;

		for (i = 0; i < msg->data_len; i++)
			p->data[7 + i] = msg->data[i];
		p->len = 7 + msg->data_len;
	} else {
		p->data[2] = msg->can_id >> 3;
		p->data[3] = (msg->can_id & 0x07) << 5;

		p->data[4] = 0;
		if (msg->is_remote_frame)
			p->data[4] |= 0x40;
		p->data[4] |= msg->data_len;

		for (i = 0; i < msg->data_len; i++)
			p->data[5 + i] = msg->data[i];
		p->len = 5 + msg->data_len;
	}
}

struct ics_interface *
ics_interface_init(const char *pname, char *err, int err_len)
{
	struct ics_interface *iface;
	struct termios t;
	struct timeval tv;
	struct ics_packet p;
	fd_set rfds;
	int com_state, determined_type, fd, i, j, is_valid;

	fd = open(pname, O_RDWR);
	if (fd == -1) {
		snprintf(err, err_len, "Could not open '%s': %s", pname,
		    strerror(errno));
		return (NULL);
	}

	if (tcgetattr(fd, &t) == -1) {
		snprintf(err, err_len, "tcgetattr failed: %s",
		    strerror(errno));
		close(fd);
		return (NULL);
	}
	cfmakeraw(&t);

	/*
	 * Apparently, POSIX didn't specify how CTS/RTS flow control was to
	 * be set, so this is a porting concern.
	 */
	#ifdef linux
	t.c_cflag |= CRTSCTS; 
	#else
	t.c_cflag |= CCTS_OFLOW | CRTS_IFLOW;
	#endif

	if (tcsetattr(fd, TCSANOW, &t) == -1) {
		snprintf(err, err_len, "tcsetattr failed: %s",
		    strerror(errno));
		close(fd);
		return (NULL);
	}

	/*
	 * XXX
	 * I doubt the following is very portable.  This implementation
	 * works under FreeBSD.  We are trying to:
	 *	clear DTR
	 *	set RTS
	 *	set DTR
	 * and then wait 250ms
	 */
	if (ioctl(fd, TIOCMGET, &com_state) == -1) {
		snprintf(err, err_len, "ioctl failed: %s", strerror(errno));
		close(fd);
		return (NULL);
	}

	com_state &= ~TIOCM_DTR;
	if (ioctl(fd, TIOCMSET, &com_state) == -1) {
		snprintf(err, err_len, "ioctl failed: %s", strerror(errno));
		close(fd);
		return (NULL);
	}

	com_state |= TIOCM_RTS;
	if (ioctl(fd, TIOCMSET, &com_state) == -1) {
		snprintf(err, err_len, "ioctl failed: %s", strerror(errno));
		close(fd);
		return (NULL);
	}

	com_state |= TIOCM_DTR;
	if (ioctl(fd, TIOCMSET, &com_state) == -1) {
		snprintf(err, err_len, "ioctl failed: %s", strerror(errno));
		close(fd);
		return (NULL);
	}

	usleep(250000);

	iface = malloc(sizeof(struct ics_interface));
	if (iface == NULL) {
		snprintf(err, err_len, "Out of memory");
		close(fd);
		return (NULL);
	}

	iface->fd = fd;
	iface->drain = iface->buf;
	iface->fill = iface->buf;
	iface->count = 0;
	iface->err[0] = '\0';

	determined_type = 0;

	for (i = 0; i < NumInterfaces && !determined_type; i++) {
		/*
		 * We're going to try and determine the adapter type by
		 * using different baud rates.  If we see a status message
		 * at 2000000 baud, then we have a ValueCAN.  If we see the
		 * status message at 3000000 baud, then we have a neoVI.
		 */
		if (tcgetattr(fd, &t) == -1) {
			snprintf(err, err_len, "tcgetattr failed: %s",
			    strerror(errno));
			close(fd);
			free(iface);
			return (NULL);
		}

		cfsetspeed(&t, Interfaces[i].baud);
		if (tcsetattr(fd, TCSANOW, &t) == -1) {
			snprintf(err, err_len, "tcsetattr failed: %s",
			    strerror(errno));
			close(fd);
			free(iface);
			return (NULL);
		}

		/* Wait up to 0.5s for a status message to appear. */
		tv.tv_sec = 0;
		tv.tv_usec = 500000;
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		if (select(fd + 1, &rfds, NULL, NULL, &tv) == -1) {
			snprintf(err, err_len, "select failed: %s",
			    strerror(errno));
			close(fd);
			free(iface);
			return (NULL);
		}
		if (FD_ISSET(fd, &rfds)) {
			/*
			 * Some data arrived.  Try read packets until we
			 * get a valid one.
			 *
			 * XXX - This is a potential dead lock if the
			 * adapter is spewing garbage, so only try 20 times.
			 */
			for (j = 0; j < 20; j++) {
				if (!ics_interface_read_packet(iface, &p,
				    &is_valid)) {
					/* read error, try next baud */
					j = 20;
					break;
				}

				if (is_valid)
					/*
					 * We don't really care what the
					 * packet is.  The fact that it
					 * was valid is enough to know
					 * we've got the correct baud rate.
					 */
					break;
			}

			if (j != 20) {
				/*
				 * We have the proper baud rate and know
				 * the adapter type.
				 */
				determined_type = 1;
				iface->index = i;
			}
		}
	}

        if (!determined_type) {
               close(fd);
               free(iface);
               snprintf(err, err_len, "could not determine interface type");
               return (NULL);
        }

        return (iface);

	#if 0
	/*
	 * We're going to check and see if Intrepid's VID is listed for any
	 * of the attached USB devices - if so, we'll use the corresponding
	 * PID and set our baud rate from that. We'll probably want to add
	 * a check for a NeoVI connected via serial (which uses 56K for
	 * comms)
	 */

	int listfd, pid;
	char usbdevs[8192];

	if ((listfd = open("/proc/bus/usb/devices", O_RDONLY)) == -1) {
		snprintf(err, err_len, "Couldn't open USB device list: %s",
		    strerror(errno));
		close(listfd);
		close(fd);
		free(iface);
		return (NULL);
	}

	if(read(listfd, usbdevs, 8192) == -1) {
		snprintf(err, err_len, "Couldn't read USB devices: %s",
		    strerror(errno));
		close(listfd);
		close(fd);
		free(iface);
		return (NULL);
	}
	
	close(listfd);

	/*
	 * Parse the PID as a hexadecimal number - 19 is the length of the
	 * preceding text
	 */
	pid = strtol(strstr(usbdevs, "Vendor=093c ProdID=") + 19, (char **)NULL, 16); 

	if (pid == 0x0601) {
		printf("vcan\n");
		iface->index = 0; /* ValueCAN */
		determined_type = 1;
	}
	else if (pid == 0x0701) {
		printf("nvi\n");
		iface->index = 1; /* NeoVI */
		determined_type = 1;
	}
	else {
		printf("wtf\n");
	/*if (!determined_type) { */
		close(fd);
		free(iface);
		snprintf(err, err_len, "could not determine interface type");
		return (NULL);
	}

	/* Set the baud rate */
	if (tcgetattr(fd, &t) == -1) {
		snprintf(err, err_len, "tcgetattr failed: %s",
		    strerror(errno));
		close(fd);
		free(iface);
		return (NULL);
	}

	cfsetspeed(&t, Interfaces[iface->index].baud);
	if (tcsetattr(fd, TCSANOW, &t) == -1) {
		snprintf(err, err_len, "tcsetattr failed: %s",
		    strerror(errno));
		close(fd);
		free(iface);
		return (NULL);
	}

	/* toss any junk */
	ics_interface_read_packet(iface, &p, &is_valid);

#endif	
	return (iface);
}

void
ics_interface_release(struct ics_interface *iface)
{

	close(iface->fd);
	free(iface);
}

int
ics_interface_fd(struct ics_interface *iface)
{

	return (iface->fd);
}

const char *
ics_interface_err_str(struct ics_interface *iface)
{

	return (iface->err);
}

static const char *
get_net_id_str(uint8_t id)
{
	const char *name;

	switch (id) {
	case ICS_NETID_DEVICE:
		name = "NETID_DEVICE";
		break;
	case ICS_NETID_HSCAN:
		name = "NETID_HSCAN";
		break;
	case ICS_NETID_MSCAN:
		name = "NETID_MSCAN";
		break;
	case ICS_NETID_SWCAN:
		name = "NETID_SWCAN";
		break;
	case ICS_NETID_LSFTCAN:
		name = "NETID_LSFTCAN";
		break;
	case ICS_NETID_FORDSCP:
		name = "NETID_FORDSCP";
		break;
	case ICS_NETID_J1708:
		name = "NETID_J1708";
		break;
	case ICS_NETID_AUX:
		name = "NETID_AUX";
		break;
	case ICS_NETID_JVPW:
		name = "NETID_JVPW";
		break;
	case ICS_NETID_ISO:
		name = "NETID_ISO";
		break;
	case ICS_NETID_ISOPIC:
		name = "NETID_ISOPIC";
		break;
	case ICS_NETID_MAIN51:
		name = "NETID_MAIN51";
		break;
	case ICS_NETID_HOST:
		name = "NETID_HOST";
		break;
	}
	return (name);
}

static ssize_t
interface_write(struct ics_interface *iface, uint8_t *ptr, size_t nbytes)
{

	return (write(iface->fd, ptr, nbytes));
}

static ssize_t
interface_read(struct ics_interface *iface, uint8_t *ptr, size_t nbytes)
{
	size_t num_to_read, num_from_buffer;
	ssize_t num_read, ret;
	uint8_t *buf_end;

	/* first try to read out of our buffer */
	num_from_buffer = interface_read_from_buffer(iface, ptr, nbytes);
	nbytes -= num_from_buffer;
	ret = num_from_buffer;

	buf_end = &iface->buf[BUFFER_SIZE];
	/* read any remaining data from file descriptor */
	while (nbytes > 0 && ret != -1) {
		/*
		 * Read as many bytes as we can hold in our buffer, even
		 * if it is more than we need.
		 */
		num_to_read = BUFFER_SIZE - iface->count;
		if (num_to_read > buf_end - iface->fill) {
			/*
			 * Don't read past end of circular buffer.  Read
			 * fewer bytes if necessary.
			 */
			num_to_read = buf_end - iface->fill;
		}
		num_read = read(iface->fd, iface->fill, num_to_read);
		if (num_read == -1)
			ret = -1;
		else {
			iface->fill += num_read;
			if (iface->fill == buf_end)
				iface->fill = iface->buf;
			iface->count += num_read;
		}

		/* now pull the data out of our buffer */
		num_from_buffer = interface_read_from_buffer(iface,
		    &ptr[ret], nbytes);
		ret += num_from_buffer;
		nbytes -= num_from_buffer;
	}
	return (ret);
}

static size_t
interface_read_from_buffer(struct ics_interface *iface, uint8_t *buf,
    size_t nbytes)
{
	size_t num_read, num_to_read, partial_read;
	uint8_t *buf_end;

	num_read = 0;
	/* first try to read out of our buffer */
	if (iface->count > 0) {
		if (nbytes > iface->count)
			num_to_read = iface->count;
		else
			num_to_read = nbytes;

		/* handle wrap around */
		buf_end = &iface->buf[BUFFER_SIZE];
		if (num_to_read <= (buf_end - iface->drain)) {
			/* we can read in one copy */
			memcpy(buf, iface->drain, num_to_read);
			iface->drain += num_to_read;
			if (iface->drain == buf_end)
				iface->drain = iface->buf;
			iface->count -= num_to_read;
		} else {
			/* the data wraps, read in two copies */
			partial_read = buf_end - iface->drain;
			memcpy(buf, iface->drain, partial_read);

			partial_read = num_to_read - partial_read;
			memcpy(&buf[partial_read], iface->buf, partial_read);
			iface->drain = &iface->buf[partial_read];
			iface->count -= num_to_read;
		}

		num_read = num_to_read;
	}
	return (num_read);
}

static int
stream_allow_can_msg(struct ics_stream_state *ss, struct ics_can_msg *msg)
{
	int allow;
	struct can_filter *f;

	/* If no filters are specified, it means pass everything */
	if (ss->filters == NULL)
		allow = 1;
	else {
		allow = 0;
		for (f = ss->filters; f != NULL; f = f->next) {
			if ((f->mask & msg->can_id) == f->masked_id) {
				allow = 1;
				break;
			}
		}
	}
	return (allow);
}
